@@ -1,11 +1,31 @@
+import itertools
 import typing
 from uuid import UUID
 
+from intervaltree import IntervalTree
+from sortedcontainers import SortedDict
+
 from .block import ByteBlock, CodeBlock, DataBlock
-from .node import Node
+from .node import Node, _NodeMessage
 from .proto import ByteInterval_pb2, SymbolicExpression_pb2
 from .symbolicexpression import SymAddrAddr, SymAddrConst, SymbolicExpression
-from .util import DictLike, SetWrapper, get_desired_range, nodes_at, nodes_on
+from .util import (
+    DictLike,
+    DictWrapper,
+    SetWrapper,
+    _IndexedAttribute,
+    _nodes_at_interval_tree,
+    _nodes_on_interval_tree,
+    _offset_interval,
+    get_desired_range,
+)
+
+if typing.TYPE_CHECKING:  # pragma: no cover
+    # Ignore flake8 "imported but unused" errors.
+    from .ir import IR  # noqa: F401
+    from .module import Module  # noqa: F401
+    from .section import Section  # noqa: F401
+
 
 SymbolicExpressionElement = typing.Tuple[
     "ByteInterval", int, SymbolicExpression
@@ -42,38 +62,80 @@ class ByteInterval(Node):
         :class:`SymbolicExpression` in the interval.
     """
 
-    class _BlockSet(SetWrapper):
-        def __init__(self, node, *args):
-            self._node = node  # type: ByteInterval
-            super().__init__(*args)
+    class _BlockSet(SetWrapper[ByteBlock]):
+        def __init__(
+            self, node: "ByteInterval", *args: typing.Iterable[ByteBlock]
+        ):
+            super().__init__()
+            self._node = node
+            self.update(*args)
 
-        def add(self, v):
-            if v._byte_interval is not None:
-                v._byte_interval.blocks.discard(v)
-            v._byte_interval = self._node
-            if self._node.ir is not None:
-                v._add_to_uuid_cache(self._node.ir._local_uuid_cache)
-            return super().add(v)
+        def add(self, v: ByteBlock) -> None:
+            # We're defining add in terms of update so that we can optimize
+            # adding multiple blocks at once.
+            self.update((v,))
 
-        def discard(self, v):
+        def update(self, *iterables: typing.Iterable[ByteBlock]) -> None:
+            node_ir = self._node.ir
+            new_items = set(*iterables) - self._data
+            for v in new_items:
+                if v._byte_interval is not None:
+                    v._byte_interval.blocks.discard(v)
+                v._byte_interval = self._node
+                if node_ir is not None:
+                    v._add_to_uuid_cache(node_ir._local_uuid_cache)
+
+            self._node._index_add_multiple(self._data, new_items)
+            self._data.update(new_items)
+
+        def discard(self, v: ByteBlock) -> None:
             if v not in self:
                 return
+            self._node._index_discard(v)
             v._byte_interval = None
             if self._node.ir is not None:
                 v._remove_from_uuid_cache(self._node.ir._local_uuid_cache)
             return super().discard(v)
 
+    class _SymbolicExprDict(DictWrapper[int, SymbolicExpression]):
+        def __init__(
+            self,
+            interval: "ByteInterval",
+            *args: DictLike[int, SymbolicExpression],
+        ):
+            # Do not call the DictWrapper constructor: _SymbolicExprDict needs
+            # to store data in a SortedDict, not a regular dict.
+            self._interval = interval
+            self._data: "SortedDict[int, SymbolicExpression]" = SortedDict()
+            temp: typing.Dict[int, SymbolicExpression] = dict(*args)
+            for i, v in temp.items():
+                self[i] = v
+
+        def __repr__(self) -> str:
+            # We can't just return the repr of self._data because it will
+            # create a SortedDict and that isn't part of our public interface.
+            items = (
+                "{!r}: {!r}".format(key, value)
+                for key, value in self._data.items()
+            )
+            return "{" + ", ".join(items) + "}"
+
+    address = _IndexedAttribute[typing.Optional[int]]()(
+        lambda self: self.section
+    )
+    size = _IndexedAttribute[int]()(lambda self: self.section)
+
     def __init__(
         self,
         *,
-        address=None,  # type: typing.Optional[int]
-        size=None,  # type: typing.Optional[int]
-        initialized_size=None,  # type: typing.Optional[int]
-        contents=b"",  # type: typing.ByteString
-        blocks=(),  # type: typing.Iterable[ByteBlock]
-        symbolic_expressions={},  # type: DictLike[int, SymbolicExpression]
-        uuid=None,  # type: typing.Optional[UUID]
-        section=None  # type: typing.Optional["Section"]
+        address: typing.Optional[int] = None,
+        size: typing.Optional[int] = None,
+        initialized_size: typing.Optional[int] = None,
+        contents: typing.ByteString = b"",
+        blocks: typing.Iterable[ByteBlock] = (),
+        symbolic_expressions: DictLike[int, SymbolicExpression] = {},
+        uuid: typing.Optional[UUID] = None,
+        section: typing.Optional["Section"] = None,
     ):
         """
         :param address: The fixed address of this interval, if present.
@@ -98,27 +160,47 @@ class ByteInterval(Node):
             raise ValueError("initialized_size must be <= size!")
 
         super().__init__(uuid=uuid)
-        self._section = None  # type: typing.Optional["Section"]
-        self.address = address  # type: typing.Optional[int]
-        self.size = size  # type: int
-        self.contents = bytearray(contents)  # type: bytearray
+        self._interval_tree: "IntervalTree[int, ByteBlock]" = IntervalTree()
+        self._section: typing.Optional["Section"] = None
+        self.address = address
+        self.size = size
+        self.contents = bytearray(contents)
         self.initialized_size = initialized_size
-        self.blocks = ByteInterval._BlockSet(
+        self.blocks: SetWrapper[ByteBlock] = ByteInterval._BlockSet(
             self, blocks
-        )  # type: typing.Set[ByteBlock]
-        self.symbolic_expressions = dict(
-            symbolic_expressions
-        )  # type: typing.Dict[int, SymbolicExpression]
-        self._proto_interval = (
-            None
-        )  # type: typing.Optional[ByteInterval_pb2.ByteInterval]
+        )
+        self._symbolic_expressions = ByteInterval._SymbolicExprDict(
+            self, symbolic_expressions
+        )
+        self._proto_interval: typing.Optional[
+            ByteInterval_pb2.ByteInterval
+        ] = (None)
 
         # Use the property setter to ensure correct invariants.
         self.section = section
 
+    def _index_add_multiple(
+        self,
+        old_blocks: typing.Collection[ByteBlock],
+        new_blocks: typing.Collection[ByteBlock],
+    ) -> None:
+        if len(old_blocks) < len(new_blocks):
+            self._interval_tree = IntervalTree(
+                _offset_interval(block)
+                for block in itertools.chain(old_blocks, new_blocks)
+            )
+        else:
+            for block in new_blocks:
+                self._index_add(block)
+
+    def _index_add(self, block: ByteBlock) -> None:
+        self._interval_tree.add(_offset_interval(block))
+
+    def _index_discard(self, block: ByteBlock) -> None:
+        self._interval_tree.discard(_offset_interval(block))
+
     @property
-    def initialized_size(self):
-        # type: () -> int
+    def initialized_size(self) -> int:
         """The number of initialized bytes in this interval.
 
         Not all bytes in this interval may correspond to bytes physically
@@ -134,8 +216,7 @@ class ByteInterval(Node):
         return len(self.contents)
 
     @initialized_size.setter
-    def initialized_size(self, value):
-        # type: (int) -> None
+    def initialized_size(self, value: int) -> None:
         if value > len(self.contents):
             self.contents += b"\0" * (value - len(self.contents))
         elif value < len(self.contents):
@@ -144,13 +225,15 @@ class ByteInterval(Node):
     @classmethod
     def _decode_protobuf(
         cls,
-        proto_interval,  # type: ByteInterval_pb2.ByteInterval
-        uuid,  # type: UUID
-        ir,  # type: typing.Optional["IR"]
-    ):
-        # type: (...) -> ByteInterval
+        proto_interval: _NodeMessage,
+        uuid: UUID,
+        ir: typing.Optional["IR"],
+    ) -> "ByteInterval":
+        assert ir
+        assert isinstance(proto_interval, ByteInterval_pb2.ByteInterval)
 
-        def decode_block(proto_block):
+        def decode_block(proto_block: ByteInterval_pb2.Block) -> ByteBlock:
+            block: ByteBlock
             if proto_block.HasField("code"):
                 block = CodeBlock._from_protobuf(proto_block.code, ir)
             elif proto_block.HasField("data"):
@@ -173,21 +256,24 @@ class ByteInterval(Node):
             size=proto_interval.size,
             contents=proto_interval.contents,
             uuid=uuid,
+            blocks=(decode_block(b) for b in proto_interval.blocks),
         )
         result._add_to_uuid_cache(ir._local_uuid_cache)
-        result.blocks.update(decode_block(b) for b in proto_interval.blocks)
         # We store the interval and IR here so we can use it later, when
         # _decode_symbolic_expressions is called.
         result._proto_interval = proto_interval
         # Return the new BI.
         return result
 
-    def _decode_symbolic_expressions(self, ir):
+    def _decode_symbolic_expressions(self, ir: "IR") -> None:
         """Called by modules after symbols are decoded, but before the module
         is done decoding.
         """
+        assert self._proto_interval
 
-        def decode_symbolic_expression(proto_expr):
+        def decode_symbolic_expression(
+            proto_expr: SymbolicExpression_pb2.SymbolicExpression,
+        ) -> SymbolicExpression:
             if proto_expr.HasField("addr_const"):
                 return SymAddrConst._from_protobuf(
                     proto_expr.addr_const, ir.get_by_uuid
@@ -212,8 +298,7 @@ class ByteInterval(Node):
 
         del self._proto_interval
 
-    def _to_protobuf(self):
-        # type: () -> ByteInterval_pb2.ByteInterval
+    def _to_protobuf(self) -> ByteInterval_pb2.ByteInterval:
         proto_interval = ByteInterval_pb2.ByteInterval()
 
         proto_interval.uuid = self.uuid.bytes
@@ -229,7 +314,7 @@ class ByteInterval(Node):
         # because append() isn't supported in older versions of protobuf. Use a
         # comprehension and extend() instead.
 
-        def to_proto_block(block):
+        def to_proto_block(block: ByteBlock) -> ByteInterval_pb2.Block:
             proto_block = ByteInterval_pb2.Block()
             proto_block.offset = block.offset
             if isinstance(block, CodeBlock):
@@ -260,22 +345,32 @@ class ByteInterval(Node):
         return proto_interval
 
     @property
-    def section(self):
-        # type: () -> typing.Optional["Section"]
+    def section(self) -> typing.Optional["Section"]:
         """The :class:`Section` this interval belongs to."""
 
         return self._section
 
     @section.setter
-    def section(self, value):
-        # type: (typing.Optional["Section"]) -> None
+    def section(self, value: typing.Optional["Section"]) -> None:
         if self._section is not None:
             self._section.byte_intervals.discard(self)
         if value is not None:
             value.byte_intervals.add(self)
 
-    def deep_eq(self, other):
-        # type: (typing.Any) -> bool
+    @property
+    def symbolic_expressions(
+        self,
+    ) -> typing.MutableMapping[int, SymbolicExpression]:
+        return self._symbolic_expressions
+
+    @symbolic_expressions.setter
+    def symbolic_expressions(
+        self, value: typing.Dict[int, SymbolicExpression]
+    ) -> None:
+        self._symbolic_expressions.clear()
+        self._symbolic_expressions.update(value)
+
+    def deep_eq(self, other: object) -> bool:
         # Do not move __eq__. See docstring for Node.deep_eq for more info.
         if not isinstance(other, ByteInterval):
             return False
@@ -308,8 +403,8 @@ class ByteInterval(Node):
             )
         )
 
-    def __repr__(self):
-        # type: () -> str
+    def __repr__(self) -> str:
+
         return (
             "ByteInterval("
             "uuid={uuid!r}, "
@@ -318,73 +413,103 @@ class ByteInterval(Node):
             "contents={contents!r}, "
             "blocks={blocks!r}, "
             "symbolic_expressions={symbolic_expressions!r}, "
-            ")".format(**self.__dict__)
+            ")".format(
+                uuid=self.uuid,
+                address=self.address,
+                size=self.size,
+                contents=self.contents,
+                blocks=self.blocks,
+                symbolic_expressions=self.symbolic_expressions,
+            )
         )
 
-    def byte_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
+    def byte_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteBlock]:
         """Finds all the byte blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.byte_blocks, addrs)
+        if self.address is None:
+            return ()
 
-    def byte_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
+        return _nodes_on_interval_tree(
+            self._interval_tree, addrs, -self.address
+        )
+
+    def byte_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteBlock]:
         """Finds all the byte blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.byte_blocks, addrs)
+        if self.address is None:
+            return ()
 
-    def code_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
+        return _nodes_at_interval_tree(
+            self._interval_tree, addrs, -self.address
+        )
+
+    def code_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[CodeBlock]:
         """Finds all the code blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.code_blocks, addrs)
+        return (
+            b for b in self.byte_blocks_on(addrs) if isinstance(b, CodeBlock)
+        )
 
-    def code_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
+    def code_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[CodeBlock]:
         """Finds all the code blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.code_blocks, addrs)
+        return (
+            b for b in self.byte_blocks_at(addrs) if isinstance(b, CodeBlock)
+        )
 
-    def data_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
+    def data_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[DataBlock]:
         """Finds all the data blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.data_blocks, addrs)
+        return (
+            b for b in self.byte_blocks_on(addrs) if isinstance(b, DataBlock)
+        )
 
-    def data_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
+    def data_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[DataBlock]:
         """Finds all the data blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.data_blocks, addrs)
+        return (
+            b for b in self.byte_blocks_at(addrs) if isinstance(b, DataBlock)
+        )
 
     def symbolic_expressions_at(
-        self, addrs  # type: typing.Union[int, range]
-    ):
-        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[SymbolicExpressionElement]:
         """Finds all the symbolic expressions that begin at an address or
         range of addresses.
 
@@ -397,20 +522,22 @@ class ByteInterval(Node):
             return
 
         addrs = get_desired_range(addrs)
-        for i, v in self.symbolic_expressions.items():
+        for i in self._symbolic_expressions._data.irange(
+            addrs.start - self.address,
+            addrs.stop - self.address,
+            inclusive=(True, False),
+        ):
             if self.address + i in addrs:
-                yield (self, i, v)
+                yield (self, i, self.symbolic_expressions[i])
 
-    def _add_to_uuid_cache(self, cache):
-        # type: (typing.Dict[UUID, Node]) -> None
+    def _add_to_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
         """Update the UUID cache when this node is added."""
 
         cache[self.uuid] = self
         for block in self.blocks:
             block._add_to_uuid_cache(cache)
 
-    def _remove_from_uuid_cache(self, cache):
-        # type: (typing.Dict[UUID, Node]) -> None
+    def _remove_from_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
         """Update the UUID cache when this node is removed."""
 
         del cache[self.uuid]
@@ -418,16 +545,14 @@ class ByteInterval(Node):
             block._remove_from_uuid_cache(cache)
 
     @property
-    def module(self):
-        # type: () -> "Module"
+    def module(self) -> typing.Optional["Module"]:
         """Get the module this node ultimately belongs to."""
         if self.section is None:
             return None
         return self.section.module
 
     @property
-    def ir(self):
-        # type: () -> "IR"
+    def ir(self) -> typing.Optional["IR"]:
         """Get the IR this node ultimately belongs to."""
         if self.module is None:
             return None
