@@ -1,3 +1,4 @@
+import collections
 import itertools
 import typing
 from enum import Enum
@@ -6,11 +7,12 @@ from uuid import UUID
 from .auxdata import AuxData, AuxDataContainer
 from .block import ByteBlock, CfgNode, CodeBlock, DataBlock, ProxyBlock
 from .byteinterval import ByteInterval, SymbolicExpressionElement
-from .node import Node
+from .node import Node, _NodeMessage
 from .proto import Module_pb2
 from .section import Section
 from .symbol import Symbol
 from .util import (
+    DeserializationError,
     DictLike,
     SetWrapper,
     nodes_at,
@@ -18,6 +20,13 @@ from .util import (
     symbolic_expressions_at,
 )
 
+if typing.TYPE_CHECKING:  # pragma: no cover
+    # Ignore flake8 "imported but unused" errors.
+    from .ir import IR  # noqa: F401
+
+
+_T = typing.TypeVar("_T", bound=typing.Union[ProxyBlock, Section, Symbol])
+
 
 class Module(AuxDataContainer):
     """Represents a loadable object, such as an executable or library.
@@ -136,24 +145,28 @@ class Module(AuxDataContainer):
         Little = Module_pb2.ByteOrder.Value("LittleEndian")
         """Little endian."""
 
-    class _NodeSet(SetWrapper):
-        def __init__(self, node, field, *args):
-            self._node = node  # type: Module
-            self._field = field  # type: str
+    class _NodeSet(SetWrapper[_T]):
+        def __init__(
+            self, node: "Module", field: str, *args: typing.Iterable[_T]
+        ):
+            self._node: Module = node
+            self._field: str = field
             super().__init__(*args)
 
-        def add(self, v):
+        def add(self, v: _T) -> None:
             if v._module is not None:
                 getattr(v._module, self._field).discard(v)
             v._module = self._node
+            self._node._index_add(v)
             if self._node.ir is not None:
                 v._add_to_uuid_cache(self._node.ir._local_uuid_cache)
             return super().add(v)
 
-        def discard(self, v):
+        def discard(self, v: _T) -> None:
             if v not in self:
                 return
             v._module = None
+            self._node._index_discard(v)
             if self._node.ir is not None:
                 v._remove_from_uuid_cache(self._node.ir._local_uuid_cache)
             return super().discard(v)
@@ -161,22 +174,21 @@ class Module(AuxDataContainer):
     def __init__(
         self,
         *,
-        name,  # type: str
-        aux_data=dict(),  # type: DictLike[str, AuxData]
-        binary_path="",  # type: str
-        file_format=FileFormat.Undefined,  # type: Module.FileFormat
-        isa=ISA.Undefined,  # type: Module.ISA
-        byte_order=ByteOrder.Undefined,  # type: Module.ByteOrder
-        preferred_addr=0,  # type: int
-        proxies=set(),  # type: typing.Iterable[ProxyBlock]
-        rebase_delta=0,  # type: int
-        sections=set(),  # type: typing.Iterable[Section]
-        symbols=set(),  # type: typing.Iterable[Symbol]
-        entry_point=None,  # type: typing.Optional[CodeBlock]
-        uuid=None,  # type: typing.Optional[UUID]
-        ir=None  # type: typing.Optional["IR"]
+        name: str,
+        aux_data: DictLike[str, AuxData] = {},
+        binary_path: str = "",
+        file_format: FileFormat = FileFormat.Undefined,
+        isa: ISA = ISA.Undefined,
+        byte_order: ByteOrder = ByteOrder.Undefined,
+        preferred_addr: int = 0,
+        proxies: typing.Iterable[ProxyBlock] = set(),
+        rebase_delta: int = 0,
+        sections: typing.Iterable[Section] = set(),
+        symbols: typing.Iterable[Symbol] = set(),
+        entry_point: typing.Optional[CodeBlock] = None,
+        uuid: typing.Optional[UUID] = None,
+        ir: typing.Optional["IR"] = None,
     ):
-        # type: (...) -> None
         """
         :param aux_data: The initial auxiliary data to be associated
             with the object, as a mapping from names to
@@ -203,24 +215,21 @@ class Module(AuxDataContainer):
         :param ir: The :class:`IR` this module belongs to.
         """
 
-        self._ir = None  # type: "IR"
-        self.binary_path = binary_path  # type: str
-        self.isa = isa  # type: Module.ISA
-        self.byte_order = byte_order  # type: Module.ByteOrder
-        self.file_format = file_format  # type: Module.FileFormat
-        self.name = name  # type: str
-        self.preferred_addr = preferred_addr  # type: int
-        self.proxies = Module._NodeSet(
-            self, "proxies", proxies
-        )  # type: typing.Set[ProxyBlock]
-        self.rebase_delta = rebase_delta  # type: int
-        self.sections = Module._NodeSet(
-            self, "sections", sections
-        )  # type: typing.Set[Section]
-        self.symbols = Module._NodeSet(
-            self, "symbols", symbols
-        )  # type: typing.Set[Symbol]
-        self.entry_point = entry_point  # type: typing.Optional[CodeBlock]
+        self._symbol_index: typing.MutableMapping[
+            str, typing.Set[Symbol]
+        ] = collections.defaultdict(set)
+        self._ir: typing.Optional["IR"] = None
+        self.binary_path = binary_path
+        self.isa = isa
+        self.byte_order = byte_order
+        self.file_format = file_format
+        self.name = name
+        self.preferred_addr = preferred_addr
+        self.proxies = Module._NodeSet(self, "proxies", proxies)
+        self.rebase_delta = rebase_delta
+        self.sections = Module._NodeSet(self, "sections", sections)
+        self.symbols = Module._NodeSet(self, "symbols", symbols)
+        self.entry_point = entry_point
         # Initialize the aux data last so that the cache is populated
         super().__init__(aux_data, uuid)
 
@@ -228,8 +237,11 @@ class Module(AuxDataContainer):
         self.ir = ir
 
     @classmethod
-    def _decode_protobuf(cls, proto_module, uuid, ir):
-        # type: (Module_pb2.Module. UUID, typing.Optiona["IR"]) -> Module
+    def _decode_protobuf(
+        cls, proto_module: _NodeMessage, uuid: UUID, ir: typing.Optional["IR"],
+    ) -> "Module":
+        assert ir
+        assert isinstance(proto_module, Module_pb2.Module)
         m = cls(
             binary_path=proto_module.binary_path,
             isa=Module.ISA(proto_module.isa),
@@ -252,11 +264,16 @@ class Module(AuxDataContainer):
             Section._from_protobuf(s, ir) for s in proto_module.sections
         )
         # entry point is a code block, which depends on sections
-        m.entry_point = (
-            ir.get_by_uuid(UUID(bytes=proto_module.entry_point))
-            if proto_module.entry_point
-            else None
-        )
+        m.entry_point = None
+        if proto_module.entry_point:
+            entry_point_uuid = UUID(bytes=proto_module.entry_point)
+            entry_point = ir.get_by_uuid(entry_point_uuid)
+            if not isinstance(entry_point, CodeBlock):
+                raise DeserializationError(
+                    "Module: entry block UUID %s is not a CodeBlock"
+                    % entry_point_uuid
+                )
+            m.entry_point = entry_point
         # symbols depend on blocks
         m.symbols.update(
             Symbol._from_protobuf(s, ir) for s in proto_module.symbols
@@ -267,15 +284,14 @@ class Module(AuxDataContainer):
                 interval._decode_symbolic_expressions(ir)
         # aux data may depend on any node
         m.aux_data.update(
-            AuxDataContainer._read_protobuf_aux_data(proto_module, ir)
+            AuxDataContainer._read_protobuf_aux_data(proto_module.aux_data, ir)
         )
 
         return m
 
-    def _to_protobuf(self):
-        # type: () -> Module_pb2.Module
+    def _to_protobuf(self) -> Module_pb2.Module:
         proto_module = Module_pb2.Module()
-        self._write_protobuf_aux_data(proto_module)
+        self._write_protobuf_aux_data(proto_module.aux_data)
         proto_module.binary_path = self.binary_path
         proto_module.isa = self.isa.value
         proto_module.file_format = self.file_format.value
@@ -291,8 +307,7 @@ class Module(AuxDataContainer):
         proto_module.uuid = self.uuid.bytes
         return proto_module
 
-    def deep_eq(self, other):
-        # type: (typing.Any) -> bool
+    def deep_eq(self, other: object) -> bool:
         # Do not move __eq__. See docstring for Node.deep_eq for more info.
         if not super().deep_eq(other):
             return False
@@ -328,8 +343,7 @@ class Module(AuxDataContainer):
 
         return True
 
-    def __repr__(self):
-        # type: () -> str
+    def __repr__(self) -> str:
         return (
             "Module("
             "uuid={uuid!r}, "
@@ -347,24 +361,42 @@ class Module(AuxDataContainer):
             ")".format(**self.__dict__)
         )
 
+    def _index_add(
+        self, node: typing.Union[ProxyBlock, Section, Symbol]
+    ) -> None:
+        if isinstance(node, Symbol):
+            self._symbol_index[node.name].add(node)
+
+    def _index_discard(
+        self, node: typing.Union[ProxyBlock, Section, Symbol]
+    ) -> None:
+        if isinstance(node, Symbol):
+            symbol_set = self._symbol_index[node.name]
+            symbol_set.discard(node)
+            if not symbol_set:
+                del self._symbol_index[node.name]
+
+    def symbols_named(self, name: str) -> typing.Iterator[Symbol]:
+        "Finds all symbols with a given name."
+        symbols = self._symbol_index.get(name, None)
+        if symbols:
+            yield from symbols
+
     @property
-    def ir(self):
-        # type: () -> typing.Optional["IR"]
+    def ir(self) -> typing.Optional["IR"]:
         """The :class:`IR` this module belongs to."""
 
         return self._ir
 
     @ir.setter
-    def ir(self, value):
-        # type: (typing.Optional["IR"]) -> None
+    def ir(self, value: typing.Optional["IR"]) -> None:
         if self._ir is not None:
             self._ir.modules.remove(self)
         if value is not None:
             value.modules.append(self)
 
     @property
-    def byte_intervals(self):
-        # type: () -> typing.Iterator[ByteInterval]
+    def byte_intervals(self) -> typing.Iterator[ByteInterval]:
         """The :class:`ByteInterval`\\s in this module."""
 
         return itertools.chain.from_iterable(
@@ -372,8 +404,7 @@ class Module(AuxDataContainer):
         )
 
     @property
-    def byte_blocks(self):
-        # type: () -> typing.Iterator[ByteBlock]
+    def byte_blocks(self) -> typing.Iterator[ByteBlock]:
         """The :class:`ByteBlock`\\s in this module."""
 
         return itertools.chain.from_iterable(
@@ -381,8 +412,7 @@ class Module(AuxDataContainer):
         )
 
     @property
-    def code_blocks(self):
-        # type: () -> typing.Iterator[CodeBlock]
+    def code_blocks(self) -> typing.Iterator[CodeBlock]:
         """The :class:`CodeBlock`\\s in this module."""
 
         return itertools.chain.from_iterable(
@@ -390,8 +420,7 @@ class Module(AuxDataContainer):
         )
 
     @property
-    def data_blocks(self):
-        # type: () -> typing.Iterator[DataBlock]
+    def data_blocks(self) -> typing.Iterator[DataBlock]:
         """The :class:`DataBlock`\\s in this module."""
 
         return itertools.chain.from_iterable(
@@ -399,14 +428,14 @@ class Module(AuxDataContainer):
         )
 
     @property
-    def cfg_nodes(self):
-        # type: () -> typing.Iterator[CfgNode]
+    def cfg_nodes(self) -> typing.Iterator[CfgNode]:
         """The :class:`CfgNode`\\s in this module."""
 
         return itertools.chain(self.code_blocks, self.proxies)
 
-    def sections_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[Section]
+    def sections_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[Section]:
         """Finds all the sections that overlap an address or range of
         addresses.
 
@@ -415,8 +444,9 @@ class Module(AuxDataContainer):
 
         return nodes_on(self.sections, addrs)
 
-    def sections_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[Section]
+    def sections_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[Section]:
         """Finds all the sections that begin at an address or range of
         addresses.
 
@@ -425,90 +455,113 @@ class Module(AuxDataContainer):
 
         return nodes_at(self.sections, addrs)
 
-    def byte_intervals_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
+    def byte_intervals_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteInterval]:
         """Finds all the byte intervals that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.byte_intervals, addrs)
+        return itertools.chain.from_iterable(
+            s.byte_intervals_on(addrs) for s in self.sections
+        )
 
-    def byte_intervals_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteInterval]
+    def byte_intervals_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteInterval]:
         """Finds all the byte intervals that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.byte_intervals, addrs)
+        return itertools.chain.from_iterable(
+            s.byte_intervals_at(addrs) for s in self.sections
+        )
 
-    def byte_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
+    def byte_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteBlock]:
         """Finds all the byte blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.byte_blocks, addrs)
+        return itertools.chain.from_iterable(
+            s.byte_blocks_on(addrs) for s in self.sections
+        )
 
-    def byte_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[ByteBlock]
+    def byte_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[ByteBlock]:
         """Finds all the byte blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.byte_blocks, addrs)
+        return itertools.chain.from_iterable(
+            s.byte_blocks_at(addrs) for s in self.sections
+        )
 
-    def code_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
+    def code_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[CodeBlock]:
         """Finds all the code blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.code_blocks, addrs)
+        return itertools.chain.from_iterable(
+            s.code_blocks_on(addrs) for s in self.sections
+        )
 
-    def code_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[CodeBlock]
+    def code_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[CodeBlock]:
         """Finds all the code blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.code_blocks, addrs)
+        return itertools.chain.from_iterable(
+            s.code_blocks_at(addrs) for s in self.sections
+        )
 
-    def data_blocks_on(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
+    def data_blocks_on(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[DataBlock]:
         """Finds all the data blocks that overlap an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_on(self.data_blocks, addrs)
+        return itertools.chain.from_iterable(
+            s.data_blocks_on(addrs) for s in self.sections
+        )
 
-    def data_blocks_at(self, addrs):
-        # type: (typing.Union[int, range]) -> typing.Iterable[DataBlock]
+    def data_blocks_at(
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[DataBlock]:
         """Finds all the data blocks that begin at an address or range of
         addresses.
 
         :param addrs: Either a ``range`` object or a single address.
         """
 
-        return nodes_at(self.data_blocks, addrs)
+        return itertools.chain.from_iterable(
+            s.data_blocks_at(addrs) for s in self.sections
+        )
 
     def symbolic_expressions_at(
-        self, addrs  # type: typing.Union[int, range]
-    ):
-        # type: (...) -> typing.Iterable[SymbolicExpressionElement]
+        self, addrs: typing.Union[int, range]
+    ) -> typing.Iterable[SymbolicExpressionElement]:
         """Finds all the symbolic expressions that begin at an address or
         range of addresses.
 
@@ -519,8 +572,7 @@ class Module(AuxDataContainer):
 
         return symbolic_expressions_at(self.sections, addrs)
 
-    def _add_to_uuid_cache(self, cache):
-        # type: (typing.Dict[UUID, Node]) -> None
+    def _add_to_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
         """Update the UUID cache when this node is added."""
 
         cache[self.uuid] = self
@@ -531,8 +583,7 @@ class Module(AuxDataContainer):
         for symbol in self.symbols:
             symbol._add_to_uuid_cache(cache)
 
-    def _remove_from_uuid_cache(self, cache):
-        # type: (typing.Dict[UUID, Node]) -> None
+    def _remove_from_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
         """Update the UUID cache when this node is removed."""
 
         del cache[self.uuid]
