@@ -1,16 +1,26 @@
 import io
-import typing
 from re import findall
-from typing import Any, NamedTuple
+from typing import (
+    Any,
+    BinaryIO,
+    Callable,
+    Collection,
+    Dict,
+    Iterable,
+    Mapping,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Type,
+    Union,
+)
 from uuid import UUID
 
 from .node import Node
 from .offset import Offset
 
-if typing.TYPE_CHECKING:
-    CacheLookupFn = typing.Optional[
-        typing.Callable[[UUID], typing.Optional[Node]]
-    ]
+CacheLookupFn = Optional[Callable[[UUID], Optional[Node]]]
 
 
 class CodecError(Exception):
@@ -28,8 +38,7 @@ class EncodeError(CodecError):
 class TypeNameError(EncodeError):
     """A type name is malformed."""
 
-    def __init__(self, hint):
-        # type: (str) -> None
+    def __init__(self, hint: str) -> None:
         super().__init__("malformed type name: '%s'" % (hint))
 
 
@@ -40,27 +49,51 @@ class UnknownCodecError(CodecError):
     :param name: the name of the unknown codec
     """
 
-    def __init__(self, name):
-        # type: (str) -> None
+    def __init__(self, name: str) -> None:
         self.name = name
 
 
-SubtypeTree = typing.Tuple[str, typing.Iterable["SubtypeTree"]]
-"""A type hint representing a parsed serialization type name.
-A ``SubtypeTree`` is a ``tuple`` with two items: A ``str`` giving
-the name of the type and a ``tuple`` of type parameters
-(which are also ``SubtypeTree``\\s). For example, the following are all
-valid ``SubtypeTree``\\s:
+class SubtypeTree:
+    """A type hint representing a parsed serialization type name.
+    A ``SubtypeTree`` is has two items: A ``str`` giving
+    the name of the type and a Sequence of type parameters
+    (which are also ``SubtypeTree``\\s). For example, the following are all
+    valid ``SubtypeTree``\\s:
 
->>> ('string', ())
->>> ('sequence', (('UUID',()),))
->>> ('mapping', (('string', ()),('set', (('UUID', ()),))))
-"""
+    >>> SubtypeTree('string', ())
+    >>> SubtypeTree('sequence', (SubtypeTree('UUID',()),))
+    >>> SubtypeTree(
+        'mapping', (
+            SubtypeTree('string', ()),
+            SubtypeTree('set', (SubtypeTree('UUID', ()),))
+        )
+    )
+    """
+
+    def __init__(self, name: str, subtypes: Sequence["SubtypeTree"]) -> None:
+        self.name = name
+        self.subtypes = subtypes
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, SubtypeTree):
+            return self.name == other.name and self.subtypes == other.subtypes
+        if isinstance(other, tuple):
+            return (self.name, self.subtypes) == other
+        return False
 
 
-class Variant(NamedTuple):
-    index: int
-    val: Any
+class Variant:
+    # Because the Variant can contain arbitrary data, depending on the context
+    # in which it is used, it has type Any. This requires an exception to the
+    # project-wide mypy configuration that disallows Any.
+    def __init__(self, index: int, val: Any):  # type: ignore[misc]
+        self.index = index
+        self.val = val
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, Variant):
+            return self.index == other.index and self.val == other.val
+        return False
 
 
 class Codec:
@@ -68,13 +101,12 @@ class Codec:
 
     @staticmethod
     def decode(
-        raw_bytes,  # type: typing.BinaryIO
+        raw_bytes: BinaryIO,
         *,
-        serialization=None,  # type: Serialization
-        subtypes=tuple(),  # type: SubtypeTree
-        get_by_uuid=None,  # type: CacheLookupFn
-    ):
-        # type: (...) -> typing.Any
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+        get_by_uuid: CacheLookupFn = None,
+    ) -> object:
         """Decode the specified raw data into a Python object.
 
         :param raw_bytes: The BytesIO object to be decoded.
@@ -85,17 +117,16 @@ class Codec:
         :returns: A new Python object, as decoded from ``raw_bytes``.
         """
 
-        raise NotImplementedError
+        raise NotImplementedError  # pragma: no cover
 
     @staticmethod
     def encode(
-        out,  # type: typing.BinaryIO
-        item,  # type: typing.Any
+        out: BinaryIO,
+        item: object,
         *,
-        serialization=None,  # type: Serialization
-        subtypes=tuple(),  # type: SubtypeTree
-    ):
-        # type: (...) -> None
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+    ) -> None:
         """Encode an item, writing the serialized object to ``out``.
 
         :param out: A binary stream to serialize to.
@@ -105,14 +136,20 @@ class Codec:
         :param subtypes: The parsed type of this object.
         """
 
-        raise NotImplementedError
+        raise NotImplementedError  # pragma: no cover
 
 
 class MappingCodec(Codec):
     """A Codec for mapping<K,V> entries. Implemented via ``dict``."""
 
     @staticmethod
-    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
+    def decode(
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Mapping[object, object]:
         try:
             key_type, val_type = subtypes
         except (TypeError, ValueError):
@@ -128,7 +165,15 @@ class MappingCodec(Codec):
         return mapping
 
     @staticmethod
-    def encode(out, mapping, *, serialization, subtypes):
+    def encode(
+        out: BinaryIO,
+        mapping: object,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+    ) -> None:
+        if not isinstance(mapping, Mapping):
+            raise EncodeError("Mapping codec only supports Mappings")
         try:
             key_type, val_type = subtypes
         except (TypeError, ValueError):
@@ -148,8 +193,12 @@ class OffsetCodec(Codec):
 
     @staticmethod
     def decode(
-        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
-    ):
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Offset:
         if subtypes != ():
             raise DecodeError("Offset should have no subtypes")
         element_uuid = UUIDCodec.decode(raw_bytes, get_by_uuid=get_by_uuid)
@@ -158,7 +207,15 @@ class OffsetCodec(Codec):
         return Offset(element_uuid, displacement)
 
     @staticmethod
-    def encode(out, val, *, serialization=None, subtypes=tuple()):
+    def encode(
+        out: BinaryIO,
+        val: object,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+    ) -> None:
+        if not isinstance(val, Offset):
+            raise EncodeError("Offset codec only supports Offsets")
         if subtypes != ():
             raise EncodeError("Offset should have no subtypes")
         UUIDCodec.encode(out, val.element_id)
@@ -169,7 +226,13 @@ class SequenceCodec(Codec):
     """A Codec for sequence<T> entries. Implemented via ``list``."""
 
     @staticmethod
-    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
+    def decode(
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Sequence[object]:
         try:
             (subtype,) = subtypes
         except (TypeError, ValueError) as e:
@@ -183,7 +246,15 @@ class SequenceCodec(Codec):
         return sequence
 
     @staticmethod
-    def encode(out, sequence, *, serialization, subtypes):
+    def encode(
+        out: BinaryIO,
+        sequence: object,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+    ) -> None:
+        if not isinstance(sequence, Sequence):
+            raise EncodeError("Sequence codec only supports Collections")
         try:
             (subtype,) = subtypes
         except (TypeError, ValueError) as e:
@@ -197,7 +268,13 @@ class SetCodec(Codec):
     """A Codec for set<T> entries. Implemented via ``set``."""
 
     @staticmethod
-    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
+    def decode(
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Set[object]:
         try:
             (subtype,) = subtypes
         except (TypeError, ValueError) as e:
@@ -211,7 +288,15 @@ class SetCodec(Codec):
         return decoded_set
 
     @staticmethod
-    def encode(out, items, *, serialization, subtypes):
+    def encode(
+        out: BinaryIO,
+        items: object,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+    ) -> None:
+        if not isinstance(items, Collection):
+            raise EncodeError("Set codec only supports Collections")
         try:
             (subtype,) = subtypes
         except (TypeError, ValueError) as e:
@@ -225,7 +310,13 @@ class TupleCodec(Codec):
     """A Codec for tuple<...> entries. Implemented via ``tuple``."""
 
     @staticmethod
-    def decode(raw_bytes, *, serialization, subtypes, get_by_uuid=None):
+    def decode(
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Tuple[object, ...]:
         # The length of a tuple is not contained in the Protobuf
         # representation, so error checking cannot be done here.
         decoded_list = list()
@@ -236,7 +327,15 @@ class TupleCodec(Codec):
         return tuple(decoded_list)
 
     @staticmethod
-    def encode(out, items, *, serialization, subtypes):
+    def encode(
+        out: BinaryIO,
+        items: object,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree],
+    ) -> None:
+        if not isinstance(items, Collection):
+            raise EncodeError("Tuple codec only supports Collections")
         if len(items) != len(subtypes):
             raise EncodeError("length of tuple does not match subtype count")
         for item, subtype in zip(items, subtypes):
@@ -248,15 +347,27 @@ class StringCodec(Codec):
 
     @staticmethod
     def decode(
-        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
-    ):
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> str:
         if subtypes != tuple():
             raise DecodeError("string should have no subtypes")
         size = Uint64Codec.decode(raw_bytes)
         return raw_bytes.read(size).decode("utf-8")
 
     @staticmethod
-    def encode(out, val, *, serialization=None, subtypes=tuple()):
+    def encode(
+        out: BinaryIO,
+        val: object,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+    ) -> None:
+        if not isinstance(val, str):
+            raise EncodeError("String codec only supports strings")
         if subtypes != ():
             raise EncodeError("string should have no subtypes")
         Uint64Codec.encode(out, len(val))
@@ -266,15 +377,19 @@ class StringCodec(Codec):
 class IntegerCodec(Codec):
     """Generic base class for integer-based Codecs"""
 
+    typname: str
+    bytesize: int
+    signed: bool
+
     @classmethod
     def decode(
         cls,
-        raw_bytes,
+        raw_bytes: BinaryIO,
         *,
-        serialization=None,
-        subtypes=tuple(),
-        get_by_uuid=None,
-    ):
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> int:
         if subtypes != ():
             raise DecodeError(f"{cls.typname} should have no subtypes")
         return int.from_bytes(
@@ -282,7 +397,16 @@ class IntegerCodec(Codec):
         )
 
     @classmethod
-    def encode(cls, out, val, *, serialization=None, subtypes=tuple()):
+    def encode(
+        cls,
+        out: BinaryIO,
+        val: object,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+    ) -> None:
+        if not isinstance(val, int):
+            raise EncodeError("Integer codec only supports integers")
         if subtypes != ():
             raise EncodeError(f"{cls.typname} should have no subtypes")
         out.write(
@@ -364,16 +488,26 @@ class UUIDCodec(Codec):
 
     @staticmethod
     def decode(
-        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
-    ):
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Union[UUID, Node]:
         if subtypes != ():
             raise DecodeError("UUID should have no subtypes")
         uuid = UUID(bytes=raw_bytes.read(16))
-        existing_node = get_by_uuid(uuid)
+        existing_node = None if get_by_uuid is None else get_by_uuid(uuid)
         return uuid if existing_node is None else existing_node
 
     @staticmethod
-    def encode(out, val, *, serialization=None, subtypes=tuple()):
+    def encode(
+        out: BinaryIO,
+        val: object,
+        *,
+        serialization: "Serialization" = None,
+        subtypes: Sequence[SubtypeTree] = (),
+    ) -> None:
         if subtypes != ():
             raise EncodeError("UUID should have no subtypes")
         if isinstance(val, Node):
@@ -394,8 +528,12 @@ class VariantCodec(Codec):
 
     @staticmethod
     def decode(
-        raw_bytes, *, serialization=None, subtypes=tuple(), get_by_uuid=None
-    ):
+        raw_bytes: BinaryIO,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree] = (),
+        get_by_uuid: Optional[CacheLookupFn] = None,
+    ) -> Variant:
         index = int.from_bytes(
             raw_bytes.read(8), byteorder="little", signed=False
         )
@@ -405,7 +543,15 @@ class VariantCodec(Codec):
         return Variant(index, val)
 
     @staticmethod
-    def encode(out, variant, *, serialization=None, subtypes=tuple()):
+    def encode(
+        out: BinaryIO,
+        variant: object,
+        *,
+        serialization: "Serialization",
+        subtypes: Sequence[SubtypeTree] = (),
+    ) -> None:
+        if not isinstance(variant, Variant):
+            raise EncodeError("Variant codec only supports variants")
         # variant is a named tuple containg index and value
         # writing the index
         out.write(variant.index.to_bytes(8, byteorder="little"))
@@ -442,11 +588,11 @@ class Serialization:
         or overridden using this dictionary.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         """Initialize with the built-in `gtirb.serialization.Codec` subclasses.
         """
 
-        self.codecs = {
+        self.codecs: Dict[str, Type[Codec]] = {
             "Addr": Uint64Codec,
             "Offset": OffsetCodec,
             "int64_t": Int64Codec,
@@ -464,10 +610,14 @@ class Serialization:
             "uint8_t": Uint8Codec,
             "UUID": UUIDCodec,
             "variant": VariantCodec,
-        }  # type: typing.Dict[str, Codec]
+        }
 
-    def _decode_tree(self, raw_bytes, type_tree, get_by_uuid):
-        # type: (typing.BinaryIO, SubtypeTree, CacheLookupFn) -> typing.Any
+    def _decode_tree(
+        self,
+        raw_bytes: BinaryIO,
+        type_tree: SubtypeTree,
+        get_by_uuid: CacheLookupFn,
+    ) -> object:
         """Decode the data in ``raw_bytes`` given a parsed type tree.
 
         :param raw_bytes: The binary stream to read bytes from.
@@ -475,22 +625,19 @@ class Serialization:
             ``raw_bytes``.
         """
 
-        try:
-            type_name, subtypes = type_tree
-        except ValueError:
-            raise DecodeError("could not unpack type tree %s" % str(type_tree))
-        if type_name not in self.codecs:
-            raise UnknownCodecError(type_name)
-        codec = self.codecs[type_name]
+        if type_tree.name not in self.codecs:
+            raise UnknownCodecError(type_tree.name)
+        codec = self.codecs[type_tree.name]
         return codec.decode(
             raw_bytes,
             serialization=self,
-            subtypes=subtypes,
+            subtypes=type_tree.subtypes,
             get_by_uuid=get_by_uuid,
         )
 
-    def _encode_tree(self, out, val, type_tree):
-        # type: (typing.BinaryIO, typing.Any, SubtypeTree) -> None
+    def _encode_tree(
+        self, out: BinaryIO, val: object, type_tree: SubtypeTree
+    ) -> None:
         """Encode the data in ``val`` given a parsed type tree.
 
         :param out: A binary stream to write bytes to.
@@ -498,18 +645,15 @@ class Serialization:
         :param type_tree: The parsed type to encode ``val`` as.
         """
 
-        try:
-            type_name, subtypes = type_tree
-        except ValueError:
-            raise EncodeError("could not unpack type tree %s" % str(type_tree))
-        if type_name not in self.codecs:
-            raise UnknownCodecError(type_name)
-        codec = self.codecs[type_name]
-        return codec.encode(out, val, serialization=self, subtypes=subtypes)
+        if type_tree.name not in self.codecs:
+            raise UnknownCodecError(type_tree.name)
+        codec = self.codecs[type_tree.name]
+        return codec.encode(
+            out, val, serialization=self, subtypes=type_tree.subtypes
+        )
 
     @staticmethod
-    def _parse_type(type_name):
-        # type: (str) -> SubtypeTree
+    def _parse_type(type_name: str) -> SubtypeTree:
         """Given an encoded aux_data type_name, generate its parse tree.
 
         >>> _parse_type('foo')
@@ -525,7 +669,9 @@ class Serialization:
         """
         tokens = findall("[^<>,]+|<|>|,", type_name)
 
-        def parse(tokens, tree):
+        def parse(
+            tokens: Sequence[str], tree: Iterable[SubtypeTree],
+        ) -> Tuple[Tuple[SubtypeTree, ...], Sequence[None]]:
             tree = list(tree)
             # It is an error to parse nothing
             if len(tokens) == 0:
@@ -538,13 +684,13 @@ class Serialization:
 
             # Base case
             if len(tail) == 0:
-                tree.append((first_token, ()))
+                tree.append(SubtypeTree(first_token, ()))
                 return tuple(tree), []
             next_token, *tail = tail
 
             # No subtypes
             if next_token == ",":
-                tree.append((first_token, ()))
+                tree.append(SubtypeTree(first_token, ()))
 
             # Parse subtypes
             if next_token == "<":
@@ -567,7 +713,7 @@ class Serialization:
                 # Parsing should consume all subtype tokens
                 if len(remaining) != 0:
                     raise TypeNameError(type_name)
-                tree.append((first_token, subtypes))
+                tree.append(SubtypeTree(first_token, subtypes))
                 # Finish if all tokens are consumed
                 if len(remaining_tokens) == 0:
                     return tuple(tree), []
@@ -589,11 +735,10 @@ class Serialization:
 
     def decode(
         self,
-        raw_bytes,  # type: typing.BinaryIO
-        type_name,  # type: str
-        get_by_uuid=None,  # type: CacheLookupFn
-    ):
-        # type: (...) -> typing.Any
+        raw_bytes: Union[bytes, bytearray, memoryview, BinaryIO],
+        type_name: str,
+        get_by_uuid: CacheLookupFn = None,
+    ) -> object:
         """Decode a :class:`gtirb.AuxData` of the specified type
         from the specified byte stream.
 
@@ -618,8 +763,7 @@ class Serialization:
             # parsed; return a blob of bytes
             return UnknownData(all_bytes)
 
-    def encode(self, out, val, type_name):
-        # type: (typing.BinaryIO,typing.Any,str) -> None
+    def encode(self, out: BinaryIO, val: object, type_name: str) -> None:
         """Encodes the value of an AuxData value to bytes.
 
         :param out: A binary stream to write bytes to.
